//----------------------------------------------------------------------------
//
// File:        tms9918a.hpp
// Date:        23-Feb-1998
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 1998-2001 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#ifndef _TMS9918A_HPP_
#define _TMS9918A_HPP_

#if defined ( ADDRESS )
    #undef ADDRESS
#endif
typedef unsigned short ADDRESS;

class cTI994A;

#define TI_TRANSPARENT		0x00
#define TI_BLACK		0x01
#define TI_MEDIUM_GREEN		0x02
#define TI_LIGHT_GREEN		0x03
#define TI_DARK_BLUE		0x04
#define TI_LIGHT_BLUE		0x05
#define TI_DARK_RED		0x06
#define TI_CYAN			0x07
#define TI_MEDIUM_RED		0x08
#define TI_LIGHT_RED		0x09
#define TI_DARK_YELLOW		0x0A
#define TI_LIGHT_YELLOW		0x0B
#define TI_DARK_GREEN		0x0C
#define TI_MAGENTA		0x0D
#define TI_GRAY			0x0E
#define TI_WHITE		0x0F

#define VDP_TIMER		0L
#define VPD_INTERRUPT_INTERVAL	17		// 60 Hz (16.666 msec)

#define VDP_WIDTH		256
#define VDP_HEIGHT		192

// VDP Register 0
#define VDP_MODE_3_BIT		0x02
#define VDP_EXTERNAL_VIDEO_MASK	0x01

// VDP Register 1
#define VDP_16K_MASK		0x80
#define VDP_BLANK_MASK		0x40
#define VDP_INTERRUPT_MASK	0x20
#define VDP_MODE_1_BIT		0x10
#define VDP_MODE_2_BIT		0x08
#define VDP_SPRITE_SIZE		0x02
#define VDP_SPRITE_MAGNIFY	0x01

#define VDP_M3			0x01	// BitMapped
#define VDP_M2			0x02	// Multi-Color
#define VDP_M1			0x04	// Text

struct sVDP_Status {
    bool  InterruptFlag   : 1;
    bool  CoincidenceFlag : 1;
    bool  FifthSpriteFlag : 1;
    UCHAR FifthSpriteNo   : 5;
};

struct sScreenImage {
    union {
        UCHAR                grapPos [ 24 ][ 32 ];
        UCHAR                textPos [ 24 ][ 40 ];
        UCHAR                data [ 0x0400 ];
    };
};

struct sColorTableEntry {
    UCHAR                    background : 4;
    UCHAR                    foreground : 4;
};

struct sColorTable {
    sColorTableEntry         data [64];
};

struct sPatternDescriptor {
    UCHAR                    data [256][8];
};

struct sSpriteAttributeEntry {
    UCHAR                    posY;
    UCHAR                    posX;
    UCHAR                    patternIndex;
    UCHAR                    earlyClock;
};

struct sSpriteAttribute {
    sSpriteAttributeEntry    data [32];
};

struct sSpriteDescriptor {
    UCHAR                    data [256][8];
};

class cTMS9918A {

protected:

    UCHAR              *m_Memory;

    sScreenImage       *m_ImageTable;
    sColorTable        *m_ColorTable;
    sPatternDescriptor *m_PatternTable;
    sSpriteAttribute   *m_SpriteAttrTable;
    sSpriteDescriptor  *m_SpriteDescTable;

    int                 m_ImageTableSize;
    int                 m_ColorTableSize;
    int                 m_PatternTableSize;

    cTI994A            *m_Host;

    ADDRESS             m_Address;
    USHORT              m_Transfer;
    USHORT              m_Shift;

    sVDP_Status         m_Status;
    UCHAR               m_Register [8];
    UCHAR               m_Mode;

    int                 m_RefreshRate;

    virtual bool SetMode ( int );

public:

    cTMS9918A ( int = 60 );
    virtual ~cTMS9918A ();

    void SetHost ( cTI994A * );

    virtual void Reset ();
    virtual void WriteAddress ( UCHAR );
    virtual void WriteData ( UCHAR );
    virtual void WriteRegister ( int, UCHAR );

    virtual UCHAR ReadData ();
    virtual UCHAR ReadRegister ( int reg )	{ return m_Register [ reg ]; }
    virtual UCHAR ReadStatus ();

    virtual bool BlankEnabled ()		{ return ( m_Register[1] & VDP_BLANK_MASK ) ? false : true; }
    virtual bool InterruptsEnabled ()		{ return ( m_Register[1] & VDP_INTERRUPT_MASK ) ? true : false; }

    virtual ADDRESS GetAddress ()		{ return ( ADDRESS ) ( m_Address & 0x3FFF ); }

    virtual void LoadImage ( FILE * );
    virtual void SaveImage ( FILE * );

    virtual void Retrace ();

    virtual bool Refresh ()			{ return false; }

    int    GetRefreshRate ()                    { return m_RefreshRate; }

    int    GetMode () const			{ return m_Mode; }
    UCHAR  *GetMemory () const			{ return m_Memory; }

    ADDRESS GetImageTable () const		{ return ( ADDRESS ) (( UCHAR * ) m_ImageTable - m_Memory ); }
    ADDRESS GetColorTable () const		{ return ( ADDRESS ) (( UCHAR * ) m_ColorTable - m_Memory ); }
    ADDRESS GetPatternTable () const		{ return ( ADDRESS ) (( UCHAR * ) m_PatternTable - m_Memory ); }
    ADDRESS GetSpriteAttrTable () const		{ return ( ADDRESS ) (( UCHAR * ) m_SpriteAttrTable - m_Memory ); }
    ADDRESS GetSpriteDescTable () const		{ return ( ADDRESS ) (( UCHAR * ) m_SpriteDescTable - m_Memory ); }

};

#endif
