//----------------------------------------------------------------------------
//
// File:        ti994a-console.cpp
// Date:        23-Feb-1998
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 1994-2001 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.hpp"
#include "compress.hpp"
#include "tms9900.hpp"
#include "tms9918a.hpp"
#include "cartridge.hpp"
#include "ti994a.hpp"
#include "ti994a-console.hpp"
#include "tms9918a-console.hpp"
#include "screenio.hpp"
#include "fileio.hpp"

extern "C" UCHAR  CpuMemory [ 0x10000 ];

extern void   Compile ( USHORT PC );
extern USHORT DisassembleASM ( USHORT, UCHAR *, char * );
extern USHORT DisassembleGPL ( USHORT, UCHAR *, char * );

#if ! defined ( PM )

static int ColumnSelect = 5;
static int KeyHead = 0;
static int KeyTail = 0;
static int KeyBuffer [ 50 ];

cConsoleTI994A::cConsoleTI994A ( cCartridge *ctg, cTMS9918A *vdp ) :
    cTI994A ( ctg, vdp )
{
}

cConsoleTI994A::~cConsoleTI994A ()
{
}

void cConsoleTI994A::KeyPressed ( int ch )
{
    ch = isupper ( ch ) ? SHIFT_KEY | ch : toupper ( ch );
    KeyBuffer [ KeyHead ] = ch;
    KeyHead = ( KeyHead + 1 ) % SIZE ( KeyBuffer );
}

int cConsoleTI994A::ReadCRU ( ADDRESS address )
{
    static UCHAR Keys[8][8] = {
        { '=',    '.', ',', 'M', 'N', '/' },		// Row 7
        { ' ',    'L', 'K', 'J', 'H', ';' },		// Row 6
        { '\x0D', 'O', 'I', 'U', 'Y', 'P' },		// Row 5
        { '\x00', '9', '8', '7', '6', '0' },		// Row 4
        { '\x00', '2', '3', '4', '5', '1' },		// Row 3
        { '\x00', 'S', 'D', 'F', 'G', 'A' },		// Row 2
        { '\x00', 'W', 'E', 'R', 'T', 'Q' },		// Row 1
        { '\x00', 'X', 'C', 'V', 'B', 'Z' } 		// Row 0
    };

    int retVal = 1;

    if (( address >= 3 ) && ( address <= 10 )) {
        if ( KeyHead == KeyTail ) return 1;
        int key = KeyBuffer [ KeyTail ];
        if ( ColumnSelect == 0 ) {
            if ( address > 5 ) {
                switch ( address ) {
                    case  7 : retVal = ( key & FCTN_KEY )  ? 0 : 1;	break;
                    case  8 : retVal = ( key & SHIFT_KEY ) ? 0 : 1;	break;
                    case  9 : retVal = ( key & CTRL_KEY )  ? 0 : 1;	break;
                }
                return retVal;
            } else if ( address == 3 ) {
                KeyTail = ( KeyTail + 1 ) % SIZE ( KeyBuffer );
            }
        }
        if ( ColumnSelect < 6 ) {
            UCHAR *row;
            switch ( address ) {
                case  3 : row = Keys [ 0 ];		break;
                case  4 : row = Keys [ 1 ];		break;
                case  5 : row = Keys [ 2 ];		break;
                case  6 : row = Keys [ 3 ];		break;
                case  7 : row = Keys [ 4 ];		break;
                case  8 : row = Keys [ 5 ];		break;
                case  9 : row = Keys [ 6 ];		break;
                case 10 : row = Keys [ 7 ];		break;
                default : fprintf ( stderr, "Invalid address (%d) for ColumSelect (%d)\n", address, ColumnSelect );
                          return 0;
            }
            if ( row [ ColumnSelect ] == ( key & 0xFF )) retVal = 0;
        }
    } else {
        retVal = cTI994A::ReadCRU ( address );
    }

    return retVal;
}

void cConsoleTI994A::WriteCRU ( ADDRESS address, USHORT data )
{
    if (( address >= 18 ) && ( address <= 20 )) {
        int shift = address - 18;
        ColumnSelect &= ~ ( 1 << shift );
        ColumnSelect |= ( data & 1 ) << shift;
    } else {
        cTI994A::WriteCRU ( address, data );
    }
}
#else
void cConsoleTI994A::KeyPressed ( int ) {}
int cConsoleTI994A::ReadCRU ( ADDRESS ) { return 1; }
void cConsoleTI994A::WriteCRU ( ADDRESS, int ) {}
#endif

#if ! defined ( PM )

#define KEY_ESCAPE  0x001B
#define KEY_UP      0x5B41
#define KEY_DOWN    0x5B42
#define KEY_RIGHT   0x5B43
#define KEY_LEFT    0x5B44

int EditNumber ( int x, int y, USHORT *value, int max, int &pos )
{
    static int hex [] = { '0', '1', '2', '3', '4', '5', '6', '7',
                   '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };
    int ch;

    ShowCursor ();

    do {
        GotoXY ( x + pos, y );
        int key = GetKey ();
        UCHAR *keyPtr = ( UCHAR * ) &key;
        ch = *keyPtr++;
        if (( ch == 0x1B ) && ( *keyPtr == 0x5B )){
            ch = ( keyPtr[0] << 8 ) | keyPtr [1];
            if ( ch == KEY_RIGHT ) { pos++;  ch = 0; }
            if ( ch == KEY_LEFT  ) { pos--;  ch = 0; }
	} else {
            ch = toupper ( ch );
            for ( unsigned i = 0; i < SIZE ( hex ); i++ ) {
                if ( ch == hex [i] ) {
                    GotoXY ( x, y );
                    if ( max == 4 ) {
                        *value = ( *value & ( 0xFFFF0FFF >> ( pos * 4 ))) | ( i << (( 3 - pos ) * 4 ));
                        outWord ( *value );
		    } else {
                        *value = ( *value & ( 0xFFFFFF0F >> ( pos * 2 ))) | ( i << (( 1 - pos ) * 2 ));
                        outByte (( UCHAR ) *value );
		    }
                    pos++;
                }
            }
            if ( ch == 0x08 ) pos--;
            if ( ch == 0x0D ) ch = KEY_DOWN;
        }
        if ( pos >= max ) ch = KEY_RIGHT;
        if ( pos < 0 ) ch = KEY_LEFT;
    } while ((( ch & 0xFF00 ) == 0 ) && ( ch != KEY_ESCAPE ));

    HideCursor ();

    return ch;
}

void cConsoleTI994A::EditRegisters ()
{
    USHORT *currReg = ( USHORT * ) &CpuMemory [ m_CPU->GetWP () ];
    int ch, pos = 0, regIndex = 0;
    do {
        ch = EditNumber ( 49 + 9 * ( regIndex >> 2 ), 5 + ( regIndex & 0x03 ), currReg + regIndex, 4, pos );
        switch ( ch ) {
            case KEY_UP    : if ( regIndex > 0 ) regIndex--;
                  	     break;
            case KEY_DOWN  : if ( regIndex < 15 ) regIndex++;
                  	     break;
            case KEY_RIGHT : if ( regIndex < 15 ) {
                                 regIndex += ( regIndex < 12 ) ? 4 : -11;
                                 pos = 0;
                             } else {
                                 pos = 3;
			     }
                             break;
            case KEY_LEFT  : if ( regIndex > 0 ) {
                                 regIndex -= ( regIndex > 3 ) ? 4 : -11;
                                 pos = 3;
                             } else {
                                 pos = 0;
			     }
                             break;
        }
    } while ( ch != KEY_ESCAPE );
}

void cConsoleTI994A::Run ()
{
    static ULONG lastClock = 0;

    Refresh ( true );
    enum { STEP, RUN };
//    ADDRESS bkptPC = 0x0070;
    ADDRESS bkptPC = 0x442C;
    int bkptCount = -1;
    int mode = STEP;
    int ch, done = false, update = 0;
    cConsoleTMS9918A *vdp = ( cConsoleTMS9918A * ) m_VDP;

    do {
        USHORT PC = m_CPU->GetPC ();

        int refresh = false;
        if (( int ) m_CPU->GetCounter () == bkptCount ) mode = STEP;
        if ( PC == bkptPC ) mode = STEP;
        if ( PC == 0x0070 ) { refresh = true; update = 0; }
        if ( update++ == 32 ) { refresh = true; update = 0; }
        if (( mode == STEP ) || refresh ) Refresh ( false );

        if ( mode == STEP ) {
            do {
                int key = GetKey ();
                UCHAR *keyPtr = ( UCHAR * ) &key;
                ch = toupper ( *keyPtr++ );
                switch ( ch ) {
                    case 'C' : bkptPC = 0xFFFF;           	        break;
                    case 'G' : mode = RUN;				break;
                    case 'R' : EditRegisters ();			break;
                    case 'V' : m_CPU->SignalInterrupt ( 1 );            break;
                    case 'B' : vdp->SetBias ( 0x60 );			break;
                    case 'N' : vdp->SetBias ( 0x00 );		 	break;
                    case 'L' : LoadImage ( "ti-994a.img" );		break;
                    case 'S' : SaveImage ( "ti-994a.img" );		break;
                }
            } while (( ch != 'G' ) && ( ch != ' ' ) && ( ch != 'Q' ));
            if ( ch == 'Q' ) break;
        }

        Compile ( PC );
        Step ();

        ULONG clockCycles = m_CPU->GetClocks ();

        // Simulate a 50/60Hz VDP interrupt
        if ( clockCycles - lastClock > m_RefreshInterval ) {
            lastClock = clockCycles;
            vdp->Retrace ();
        }

        if (( mode == RUN ) && ::KeyPressed ()) {
            int key = GetKey ();
            UCHAR *keyPtr = ( UCHAR * ) &key;
            ch = *keyPtr++;
            if ( ch == KEY_ESCAPE ) {
                if ( *keyPtr++ != 0x5B ) {
                    mode = STEP;
                    continue;
                }
                ch = 0x5B00 | *keyPtr;
            }
            KeyPressed ( ch );
        }

    } while ( ! done );
}
#endif

#if defined ( LOG_GROM )
FILE *gplFile = fopen ( "GPL.HST", "wt" );
int level = 0;
#endif
bool cConsoleTI994A::Step ()
{
    m_CPU->Step ();
    if ( m_CPU->GetPC () == 0x0070 ) {
        m_GromCounter++;
#if defined ( LOG_GROM )
        UCHAR *gromPtr = m_GromPtr;
        char *gptr = DisassembleGPL ( gromPtr );
        char *tempptr = &gptr [ strlen ( gptr ) ];
        while ( *--tempptr == ' ' );
        *++tempptr = '\0';
        fprintf ( gplFile, "%*.*s%04X %s\n", level, level, "", GromAddress, gptr );
        if ( *GromPtr == 0x06 ) level += 2;
        else if (( level > 0 ) && (( *GromPtr == 0x00 ) || ( *GromPtr == 0x01 ))) level -= 2;
#endif
    }
    return false;
}

void cConsoleTI994A::SaveImage ( const char *filename )
{
    char buffer [256];

    sprintf ( buffer, "%s%c%s", HOME_PATH, FILE_SEPERATOR, filename );

    cTI994A::SaveImage ( buffer );
}

bool cConsoleTI994A::LoadImage ( const char *filename )
{
    char buffer [256];

    sprintf ( buffer, "%s%c%s", HOME_PATH, FILE_SEPERATOR, filename );

    return cTI994A::LoadImage ( buffer );
}

USHORT cConsoleTI994A::VideoReadBreakPoint ( const ADDRESS address, USHORT data )
{
    switch ( address ) {
        case 0x8800 : data = m_VDP->ReadData ();
                      GotoXY ( 75, 1 );	outWord ( m_VDP->GetAddress ());
                      GotoXY ( 76, 2 );	outByte (( UCHAR ) data );
                      break;
        case 0x8802 : data = m_VDP->ReadStatus ();
                      break;
    }
    return data;
}

USHORT cConsoleTI994A::VideoWriteBreakPoint ( const ADDRESS address, USHORT data )
{
    switch ( address ) {
        case 0x8C00 : m_VDP->WriteData (( UCHAR ) data );
                      GotoXY ( 76, 2 );	outByte (( UCHAR ) data );
                      break;
        case 0x8C02 : m_VDP->WriteAddress (( UCHAR ) data );
                      break;
    }
    GotoXY ( 75, 1 );	outWord ( m_VDP->GetAddress ());
    return data;
}

USHORT cConsoleTI994A::GromReadBreakPoint ( const ADDRESS address, USHORT data )
{
    switch ( address ) {
        // Grom/Gram Read Byte Port
        case 0x9800 : data = *m_GromPtr++;
                      m_GromAddress++;
                      GotoXY ( 63, 1 );	outWord ( m_GromAddress );
                      GotoXY ( 64, 2 );	outByte (( UCHAR ) data );
                      break;
        // Grom/Gram Read Address Port
        case 0x9802 : data = ( USHORT ) ((( m_GromAddress + 1 ) >> m_GromReadShift ) & 0x00FF );
                      m_GromReadShift = 8 - m_GromReadShift;
                      m_GromPtr = &m_GromMemory [ m_GromAddress ];
                      m_GromWriteShift = 8;
                      break;
    }
    return data;
}

USHORT cConsoleTI994A::GromWriteBreakPoint ( const ADDRESS address, USHORT data )
{
    sMemoryRegion *memory;
    switch ( address ) {
        // Grom/Gram Write Byte Port
        case 0x9C00 : memory = m_GromMemoryInfo [ m_GromAddress >> 13 ];
                      if ( memory && ( memory->CurBank->Type != MEMORY_ROM )) *m_GromPtr = ( UCHAR ) data;
                      m_GromPtr++;
                      m_GromAddress++;
                      GotoXY ( 64, 2 );	outByte (( UCHAR ) data );
                      break;
        // Grom/Gram Write (set) Address Port
        case 0x9C02 : m_GromAddress &= ( USHORT ) ( 0xFF00 >> m_GromWriteShift );
                      m_GromAddress |= ( USHORT ) ( data << m_GromWriteShift );
                      m_GromWriteShift = 8 - m_GromWriteShift;
                      m_GromPtr = &m_GromMemory [ m_GromAddress ];
                      m_GromReadShift = 8;
                      break;
    }
    GotoXY ( 63, 1 );	outWord ( m_GromAddress );
    return data;
}

void cConsoleTI994A::Refresh ( bool complete )
{
    static ADDRESS lastWP = 0;
    static UCHAR   vdpReg [8];
    static USHORT  lastST = 0, lastReg [16];
    char buffer [ 256 ];

    if ( complete ) {
        PutXY ( 46,  1, "WP: ", 4 );
        PutXY ( 46,  2, "PC: ", 4 );
        PutXY ( 46,  3, "ST: ", 4 );

        PutXY ( 57,  1, "GROM: ", 6 );	outWord ( m_GromAddress );
        PutXY ( 57,  2, "Byte:  --", 9 );

        PutXY ( 69,  1, " VDP: ", 6 );	outWord ( m_VDP->GetAddress ());
        PutXY ( 69,  2, "Byte:  --", 9 );

        PutXY ( 46,  5, "R0 0000  R4 0000  R8 0000 R12 0000", 34 );
        PutXY ( 46,  6, "R1 0000  R5 0000  R9 0000 R13 0000", 34 );
        PutXY ( 46,  7, "R2 0000  R6 0000 R10 0000 R14 0000", 34 );
        PutXY ( 46,  8, "R3 0000  R7 0000 R11 0000 R15 0000", 34 );

        PutXY ( 46, 10, "V0 00    V1 00    V2 00    V3 00", 32 );
        PutXY ( 46, 11, "V4 00    V5 00    V6 00    V7 00", 32 );

        PutXY ( 46, 13, "GPL:", 4 );
        PutXY ( 46, 20, "CPU:", 4 );

        lastWP = ( ADDRESS ) -1;
        lastST = ( USHORT ) -1;
        memset ( lastReg, -1, sizeof ( lastReg ));
        memset ( vdpReg, -1, sizeof ( vdpReg ));
    }

    char *spaces = "                                 ";
    USHORT PC = m_CPU->GetPC ();
    if ( complete || ( PC == 0x0070 )) {
        if ( PC == 0x0070 ) m_GromLastInstruction = m_GromAddress;
        UCHAR *gromPtr = m_GromPtr;
        for ( int i = 0; i < 5; i++ ) {
            USHORT gromAddress = ( USHORT ) ( m_GromAddress + ( gromPtr  - m_GromPtr ));
            gromPtr += DisassembleGPL ( gromAddress, gromPtr, buffer );
            int len = strlen ( buffer );
            if ( len > 33 ) len = 33;
            PutXY ( 46, 14+i, buffer, len );
            if ( len != 33 ) PutXY ( 46+len, 14+i, spaces, 33-len );
        }
        GotoXY ( 72, 13 );	outLong ( m_GromCounter );
    }
    if ( complete || ( m_CPU->GetWP () != lastWP )) {
        GotoXY ( 50, 1 );	outWord ( lastWP = m_CPU->GetWP ());
    }
    GotoXY ( 50, 2 );	outWord ( PC );
    if ( complete || ( m_CPU->GetST () != lastST )) {
        GotoXY ( 50, 3 );   outWord ( lastST = m_CPU->GetST ());
    }

    GotoXY ( 72, 20 );	outLong ( m_CPU->GetCounter ());
    USHORT *currReg = ( USHORT * ) &CpuMemory [ lastWP ];

    for ( int i = 0; i < 16; i++ ) {
        if ( complete || ( lastReg[i] != currReg [i] )) {
            GotoXY ( 49 + 9 * ( i >> 2 ), 5 + ( i & 0x03 ));
            lastReg[i] = currReg [i];
            USHORT value = ( CpuMemory [ lastWP + i * 2 ] << 8 ) | CpuMemory [ lastWP + i * 2 + 1 ];
            outWord ( value );
        }
    }

    for ( int i = 0; i < 8; i++ ) {
        if ( complete || ( vdpReg[i] != m_VDP->ReadRegister ( i ))) {
            GotoXY ( 49 + ( i % 4 ) * 9, ( i < 4 ) ? 10 : 11 );
            outByte (( UCHAR ) ( vdpReg[i] = m_VDP->ReadRegister ( i )));
        }
    }

    for ( int i = 0; i < 5; i++ ) {
        PC = DisassembleASM ( PC, &CpuMemory [ PC ], buffer );
        int len = strlen ( buffer );
        if ( len > 33 ) len = 33;
        PutXY ( 46, 21+i, buffer, len );
        if ( len != 33 ) PutXY ( 46+len, 21+i, spaces, 33-len );
    }
}
